@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2022, NVIDIA CORPORATION. All rights reserved.
+ * Copyright (c) 2014-2024, NVIDIA CORPORATION. All rights reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -39,18 +39,26 @@ typedef struct WlEglStreamImageRec {
     /* Pointer back to the parent surface for use in Wayland callbacks */
     struct WlEglSurfaceRec *surface;
 
-    /*
-     * Use an individual mutex to guard access to each image's data. This avoids
-     * sharing the surface lock between the app and buffer release event
-     * threads, resulting in simplified lock management and smaller critical
-     * sections.
-     */
-    pthread_mutex_t         mutex;
-
     EGLImageKHR             eglImage;
     struct wl_buffer       *buffer;
     EGLBoolean              attached;
     struct wl_list          acquiredLink;
+
+    struct wp_linux_drm_syncobj_timeline_v1 *wlReleaseTimeline;
+    uint32_t                drmSyncobjHandle;
+    int                     releasePending;
+    /* Latest release point the compositor will signal with explicit sync */
+    uint64_t                releasePoint;
+    /* Cached acquire EGLSync from acquireImage */
+    EGLSyncKHR              acquireSync;
+
+    /*
+     * Used for delaying the destruction of the image if we are waiting the
+     * buffer release thread to use it later.
+     */
+    EGLBoolean              destructionPending;
+
+    struct wl_list          link;
 } WlEglStreamImage;
 
 typedef struct WlEglSurfaceCtxRec {
@@ -70,14 +78,16 @@ typedef struct WlEglSurfaceCtxRec {
     EGLuint64KHR framesProcessed;
 
     /*
-     * The double pointer is because of the need to allocate the data for each
-     * image slot separately to avoid clobbering the acquiredLink member
-     * whenever the streamImages arrary is resized with realloc().
+     * Use an individual mutex to guard access to streamImages. This helps us
+     * to avoid sharing the surface lock between the app and buffer release
+     * event threads, resulting in simplified lock management and smaller
+     * critical sections.
      */
-    WlEglStreamImage      **streamImages;
+    pthread_mutex_t         streamImagesMutex;
+
+    struct wl_list          streamImages;
     struct wl_list          acquiredImages;
     struct wl_buffer       *currentBuffer;
-    uint32_t                numStreamImages;
 
     struct wl_list link;
 } WlEglSurfaceCtx;
@@ -151,6 +161,13 @@ struct WlEglSurfaceRec {
     EGLBoolean isResized;
 
     WlEglDmaBufFeedback feedback;
+
+    /* per-surface Explicit Sync objects */
+    struct wp_linux_drm_syncobj_surface_v1 *wlSyncobjSurf;
+    struct wp_linux_drm_syncobj_timeline_v1 *wlAcquireTimeline;
+    uint32_t drmSyncobjHandle;
+    /* Last acquire point used. This starts at 1, zero means invalid.  */
+    uint64_t syncPoint;
 };
 
 void wlEglReallocSurface(WlEglDisplay *display,
@@ -185,6 +202,9 @@ EGLBoolean wlEglQueryNativeResourceHook(EGLDisplay dpy,
                                         EGLint attribute,
                                         int *value);
 
+EGLBoolean
+wlEglSurfaceCheckReleasePoints(WlEglDisplay *display, WlEglSurface *surface);
+
 EGLBoolean wlEglSendDamageEvent(WlEglSurface *surface,
                                 struct wl_event_queue *queue);
 
